/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

// HistogramDialog.cpp : implementation file
//

#include "stdafx.h"
#include "Lafter.h"
#include "HistogramDialog.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// Make a standard BITMAPINFO_8 structure
static BITMAPINFO_8 stdBmInfo;
TVector2f               skinMean( 0.490607f, 0.303686f);
TMatrix3f               skinCov( 0.2477e-03f,-0.0501e-03f,0.1124e-03f);
// Vasi's model
//TVector2f               skinMean( 0.470607, 0.303686);
//TMatrix3f               skinCov( 0.2477e-03,-0.0501e-03,0.1124e-03);
// This is the inverse of the covariance matrix
//TMatrix3f               skinInvCov( 4437.56,1978.06,9776.8);
TGaussianf              skinModel(skinMean,skinCov);

/////////////////////////////////////////////////////////////////////////////
// CHistogramDialog dialog

CHistogramDialog::CHistogramDialog(CWnd* pParent /*=NULL*/)
: CDialog(CHistogramDialog::IDD, pParent)
{
	//{{AFX_DATA_INIT(CHistogramDialog)
	m_fMaxDist = 0.0f;
	m_fResolution = 1000.0f;
	m_nColorSource = 1;
	m_bRobust = FALSE;
	//}}AFX_DATA_INIT
	
	
	m_nThreshold = 100;
	m_nSliderThreshold = 0;
	m_pDlg = NULL;
	m_pGrabber = NULL;
	
	m_bGrabbedImgDlg = FALSE;
	m_bFirstGrabbedImgDlg = FALSE;
	m_pGrabbedImgDlg = NULL;
	m_colorModel.set( skinModel );
	
	memset((void*)&m_stdBmInfo, 0, sizeof(m_stdBmInfo));
	::InitializeCriticalSection(&m_Sem);
}

CHistogramDialog::CHistogramDialog(CDialog* pDlg)
: CDialog(CHistogramDialog::IDD, pDlg), m_picProcessedImg( pDlg )
{
	m_fMaxDist = 50.0f;
	m_nThreshold = 0;
	m_nSliderThreshold = (255 - m_nThreshold);
	m_fResolution = 1000.0f;
	m_bRobust = FALSE;
	m_nColorSource = 1;
	m_pGrabber = NULL;
	m_colorModel.set( skinModel );
	
	m_pDlg = pDlg;
	
	m_bGrabbedImgDlg = FALSE;
	m_bFirstGrabbedImgDlg = FALSE;
	m_pGrabbedImgDlg = new CGrabbedImageDialog( this);
	::InitializeCriticalSection(&m_Sem);
}

CHistogramDialog::~CHistogramDialog() {
	::DeleteCriticalSection(&m_Sem);
	delete m_pGrabbedImgDlg;
}

void CHistogramDialog::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CHistogramDialog)
	DDX_Text(pDX, IDC_MAXDIST, m_fMaxDist);
	DDV_MinMaxFloat(pDX, m_fMaxDist, 0.f, 99999.f);
	DDX_Text(pDX, IDC_RESOLUTION, m_fResolution);
	DDV_MinMaxFloat(pDX, m_fResolution, 0.f, 99999.f);
	DDX_Radio(pDX, IDC_COLORSOURCE, m_nColorSource);
	DDX_Check(pDX, IDC_ROBUST, m_bRobust);
	DDX_Control(pDX, IDC_PROCESSEDIMG, m_picProcessedImg);
	//}}AFX_DATA_MAP
	::EnterCriticalSection( &m_Sem );
	m_histo.SetResolution( m_fResolution );
	m_histo.SetMaxDist( m_fMaxDist );
	::LeaveCriticalSection( &m_Sem );
}


BOOL CHistogramDialog::Create( int nPixSize, int nWidth, int nHeight,
							  VTciGrabber		*pGrabber,
							  int nInitThreshold,
							  float fInitMaxDist )
{
	BOOL	err	= CDialog::Create(CHistogramDialog::IDD);
	
	m_fMaxDist = fInitMaxDist;
	m_nThreshold = nInitThreshold;
	m_nSliderThreshold = (255 - m_nThreshold);
	
	CSliderCtrl* pSlide = (CSliderCtrl*) GetDlgItem(IDC_TRACK_THRESHOLD);
	CString strText;
	pSlide->SetPos(m_nSliderThreshold);
	strText.Format("%d", m_nThreshold);
	SetDlgItemText(IDC_THRESHOLD, strText);
	
	m_fResolution = 1000.0f;
	m_bRobust = FALSE;
	m_nColorSource = 1;
	m_pGrabber = pGrabber;
	
	m_nFrameWidth = nWidth;
	m_nFrameHeight = nHeight;
	
	memset((void *)&m_stdBmInfo, 0, sizeof(m_stdBmInfo));
	m_stdBmInfo.bmiHeader.biSize			= sizeof(BITMAPINFOHEADER);
	m_stdBmInfo.bmiHeader.biWidth			= m_nFrameWidth;
	m_stdBmInfo.bmiHeader.biHeight			= m_nFrameHeight;
	m_stdBmInfo.bmiHeader.biPlanes			= 1;
	m_stdBmInfo.bmiHeader.biBitCount		= nPixSize; 
	m_stdBmInfo.bmiHeader.biCompression		= BI_RGB;
    m_stdBmInfo.bmiHeader.biSizeImage		= 0;
    m_stdBmInfo.bmiHeader.biXPelsPerMeter	= 0;
    m_stdBmInfo.bmiHeader.biYPelsPerMeter	= 0;
    m_stdBmInfo.bmiHeader.biClrUsed			= 0;
    m_stdBmInfo.bmiHeader.biClrImportant	= 0;
	
	// Calculate the size of the image
	UINT wWidthBytes = (m_nFrameWidth * nPixSize);
	wWidthBytes = ((wWidthBytes+31) & (~31))/ 8;
	
	m_stdBmInfo.bmiHeader.biSizeImage = wWidthBytes * abs ((int)m_stdBmInfo.bmiHeader.biHeight);
	
	m_picProcessedImg.Create(m_stdBmInfo, FALSE);
	
	// Init the histogram object
	m_histo.Init((unsigned long)m_fResolution);
	::EnterCriticalSection( &m_Sem);
	m_histo.SetMaxDist( m_fMaxDist );
	m_histo.Compute(m_colorModel, m_nThreshold, m_fMaxDist );
	::LeaveCriticalSection(&m_Sem);
	return err;
}

BEGIN_MESSAGE_MAP(CHistogramDialog, CDialog)
//{{AFX_MSG_MAP(CHistogramDialog)
ON_WM_CLOSE()
ON_WM_VSCROLL()
ON_BN_CLICKED(IDC_COLORSOURCE, OnColorsource)
ON_BN_CLICKED(IDC_RADIO7, OnRadio7)
ON_BN_CLICKED(IDC_GRAB, OnGrab)
ON_BN_CLICKED(IDC_COMPUTE, OnCompute)
ON_BN_CLICKED(IDC_SAVE, OnSave)
ON_BN_CLICKED(IDC_SETFROMHISTO, OnSetfromhisto)
ON_BN_CLICKED(IDC_ROBUST, OnRobust)
//}}AFX_MSG_MAP

END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CHistogramDialog message handlers

void CHistogramDialog::OnClose() 
{
	// TODO: Add your message handler code here and/or call default
	m_pDlg->PostMessage(WM_HGGOODBYE, 1);
}

void CHistogramDialog::CleanUp()
{
	if (m_bFirstGrabbedImgDlg) {
		m_pGrabbedImgDlg->DestroyWindow();
		m_bFirstGrabbedImgDlg = FALSE;
	}
	
	m_histo.Dispose();
}


void CHistogramDialog::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	// TODO: Add your message handler code here and/or call default
	
	CSliderCtrl* pSlide = (CSliderCtrl*) pScrollBar;
	CString strText;
	
	switch (pScrollBar->GetDlgCtrlID()) {
	case IDC_TRACK_THRESHOLD:
		m_nSliderThreshold = pSlide->GetPos();
		m_nThreshold = m_nArrayThreshold[m_nSliderThreshold];
		strText.Format("%d", m_nThreshold);
		SetDlgItemText(IDC_THRESHOLD, strText);
		break;
	}
	
	if (m_pDlg != NULL) {
		m_pDlg->PostMessage(WM_HGUPDATE,1);
	}
}

BOOL CHistogramDialog::OnInitDialog() 
{
	CDialog::OnInitDialog();
	
	// Initialize Slider Array
	for (int i=0, j=255;i<256;i++, j--) {
		m_nArrayThreshold[i] = j;
	}
	
	m_nSliderThreshold = m_nArrayThreshold[m_nThreshold];
	
	// TODO: Add extra initialization here
	CSliderCtrl* pSlide = (CSliderCtrl*) GetDlgItem(IDC_TRACK_THRESHOLD);
	CString strText;
	pSlide->SetRange(0,255);
	pSlide->SetPos(m_nSliderThreshold);
	strText.Format("%d", m_nThreshold);
	SetDlgItemText(IDC_THRESHOLD, strText);
	
	CButton* GaussianButton = (CButton*) GetDlgItem(IDC_SETFROMHISTO);
	if (m_nColorSource == 1) {
		Show(GaussianButton);
	}
	
	m_pointGrabbedImgDlg = CPoint(600, 100);
	
	UpdateData(TRUE);
	return TRUE;  // return TRUE unless you set the focus to a control
	// EXCEPTION: OCX Property Pages should return FALSE
}

void CHistogramDialog::OnColorsource() 
{
	// TODO: Add your control notification handler code here
	UpdateData(TRUE);
	
	// Call the static function of the VDib class to set the Screen Information
	VDib::setBitmapFormat();
	
	CString WindowName("Grabbed Image");
	
	// Tempoary. Must be passed in
	BITMAPINFO_8 stdBmInfo;
	stdBmInfo.bmiHeader.biSize			= sizeof(BITMAPINFOHEADER);
	stdBmInfo.bmiHeader.biWidth			= m_pGrabber->Width();
	stdBmInfo.bmiHeader.biHeight		= m_pGrabber->Height();
	stdBmInfo.bmiHeader.biPlanes		= 1;
	stdBmInfo.bmiHeader.biBitCount		= m_pGrabber->PixelSize();
	stdBmInfo.bmiHeader.biCompression	= BI_RGB;
    stdBmInfo.bmiHeader.biSizeImage		= 0;
    stdBmInfo.bmiHeader.biXPelsPerMeter = 0;
    stdBmInfo.bmiHeader.biYPelsPerMeter = 0;
    stdBmInfo.bmiHeader.biClrUsed		= 0;
    stdBmInfo.bmiHeader.biClrImportant	= 0;
	
	// Calculate the size of the image
	UINT wWidthBytes = (stdBmInfo.bmiHeader.biWidth * stdBmInfo.bmiHeader.biBitCount);
	wWidthBytes = ((wWidthBytes+31) & (~31))/ 8;
	
	stdBmInfo.bmiHeader.biSizeImage = wWidthBytes * abs ((int)stdBmInfo.bmiHeader.biHeight);
	
	if (m_nColorSource == 0)
	{
		if (!m_bFirstGrabbedImgDlg) {
			m_pGrabbedImgDlg->Create( stdBmInfo.bmiHeader.biBitCount,
															stdBmInfo.bmiHeader.biWidth, 
															stdBmInfo.bmiHeader.biHeight, FALSE, TRUE, FALSE);
			m_pGrabbedImgDlg->SetWindowPos(&wndTop,0,400,0,0,
				SWP_HIDEWINDOW | SWP_NOSIZE);
			m_pGrabbedImgDlg->SetWindowText("Grabbed Image");
			m_pGrabbedImgDlg->CenterBitmap();
			m_bFirstGrabbedImgDlg = TRUE;
		}
		
		CButton* GaussianButton = (CButton*) GetDlgItem(IDC_SETFROMHISTO);
		CButton* GrabButton = (CButton*) GetDlgItem(IDC_GRAB);
		Hide(GaussianButton);
		Show(m_pGrabbedImgDlg);
		Show(GrabButton);
		Invalidate();
		UpdateWindow();
		m_bGrabbedImgDlg = TRUE;
	}
}

long CHistogramDialog::OnBMGoodbye(UINT wParam, long lParam)
{
	m_bGrabbedImgDlg = FALSE;
	return 0L;
}

void CHistogramDialog::OnRadio7() 
{
	// TODO: Add your control notification handler code here
	UpdateData(TRUE);
	
	if (m_nColorSource == 1)
	{
		CButton* GaussianButton = (CButton*) GetDlgItem(IDC_SETFROMHISTO);
		CButton* GrabButton = (CButton*) GetDlgItem(IDC_GRAB);
		Hide(GrabButton);
		if (m_bFirstGrabbedImgDlg) {
			Hide(m_pGrabbedImgDlg);
		}
		Show(GaussianButton);		
		Invalidate();
		UpdateWindow();
		m_bGrabbedImgDlg = FALSE;
	}
}

void CHistogramDialog::OnGrab() 
{
	// TODO: Add your control notification handler code here
	// Attach the grabbed image from the grabber to the static frame
	if ( m_bGrabbedImgDlg )
		m_pGrabber->Snap(&(m_pGrabbedImgDlg->m_picLiveImg));
}

void CHistogramDialog::OnCompute() 
{
	if ( !m_bGrabbedImgDlg ) {
		::EnterCriticalSection( &m_Sem );
		m_histo.Compute(m_colorModel, m_nThreshold, m_fMaxDist );
		::LeaveCriticalSection( &m_Sem );
	} else {
		if (!(m_pGrabbedImgDlg->m_bFirstBox)) {
			CRect		auxRect = m_pGrabbedImgDlg->GetBoundingBox();
			//Compute the histogram from the grabbed bitmap
			::EnterCriticalSection( &m_Sem );
			m_histo.Compute(&(m_pGrabbedImgDlg->m_picLiveImg.m_dib), 
				m_bRobust, 
				m_nThreshold, 
				&auxRect,
				m_fMaxDist );
			::LeaveCriticalSection( &m_Sem );
		}
	}
	// TODO: Add your control notification handler code here
	UpdateData(TRUE);
}

void CHistogramDialog::OnSave() 
{
	// TODO: Add your control notification handler code here
	CFileDialog dlg(FALSE);
	CFile DataFile;
	CFileException fileException;
	
	// If the grabbed image dialog is active, save the bounding box
	// otherwise save the gaussian
	
	if ( m_bGrabbedImgDlg ) {
		if (m_pGrabbedImgDlg->m_bFirstBox) {
		} else {
			if (dlg.DoModal() == IDOK) {
				if ( !DataFile.Open( dlg.GetPathName( ), CFile::modeCreate | CFile::modeWrite, &fileException) ) {
					TRACE( "Can't open file %s, error = %u\n",
						dlg. GetPathName( ), fileException.m_cause );
				} else {
					// storing code
					CString szBuffer;
					CString tempBuffer;
					CRect BoundingBox = m_pGrabbedImgDlg->GetBoundingBox();
					CDC* pDC = GetDC();
					pDC->DPtoLP(BoundingBox);
					ReleaseDC(pDC);
					int nWidth = BoundingBox.Width();
					int nHeight = BoundingBox.Height();
					int PixelNumBytes = m_pGrabber->PixelSize()/8;
					int lineOffset = nWidth*PixelNumBytes;
					unsigned char* linePtr = (unsigned char*) m_pGrabbedImgDlg->m_picLiveImg.m_dib.GetBits();
					// Set lpbits pointer to upper left corner of IMAGE
					linePtr += lineOffset * abs(BoundingBox.top);
					// sets pointer to left edge of BOX
					linePtr += BoundingBox.left * PixelNumBytes;
					unsigned char* lpBits = linePtr;
					unsigned char r, g, b;
					
					for (int rows=0;rows<nHeight;rows++)  {
						for (int cols=0;cols<nWidth;cols++) {
							b=*lpBits;
							lpBits++;
							g=*lpBits;
							lpBits++;
							r=*lpBits;
							lpBits+=2;
							tempBuffer.Format("%d %d %d \n", r,g,b);
							szBuffer += tempBuffer;
						}
						// Sets pointer to beginning of next line
						linePtr += lineOffset;
						lpBits = linePtr;
					}
					DataFile.Write(szBuffer, szBuffer.GetLength());
					DataFile.Close();					
				} // else
			} // if doModal
		} // if m_pGrabbedImgDlg->m_bFirstBox
	} else { // m_bGrabbedImgDlg == FALSE
		// save the gaussian info
		if (dlg.DoModal() == IDOK) {
			if ( !DataFile.Open( dlg.GetPathName( ), CFile::modeCreate | CFile::modeWrite, &fileException) ) {
				TRACE( "Can't open file %s, error = %u\n",
					dlg. GetPathName( ), fileException.m_cause );
			} else {
				// Save the gaussian in file
				m_colorModel.Write( DataFile );
				DataFile.Close();
			}
		}
	}
}	

int CHistogramDialog::Hide( )
{
	if (m_bGrabbedImgDlg) {
		Hide(m_pGrabbedImgDlg);
	}
	SetWindowPos(&wndTop,0,0,0,0, SWP_HIDEWINDOW| SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
	return 0;
}

int CHistogramDialog::Hide(CWnd* pWnd)
{
	pWnd->SetWindowPos(&wndTop,0,0,0,0, SWP_HIDEWINDOW| SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
	return 0;
}

int CHistogramDialog::Show( )
{
	if (m_bGrabbedImgDlg) {
		Show(m_pGrabbedImgDlg);
	}
	SetWindowPos(&wndTop,0,0,0,0, SWP_SHOWWINDOW| SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
	return 0;
}

int CHistogramDialog::Show(CWnd* pWnd)
{
	pWnd->SetWindowPos(&wndTop,0,0,0,0, SWP_SHOWWINDOW| SWP_NOMOVE | SWP_NOSIZE | SWP_NOZORDER);
	return 0;
}

void CHistogramDialog::OnSetfromhisto() 
{
	// TODO: Add your control notification handler code here
	m_histo.GetGaussian ( m_colorModel );
}

void CHistogramDialog::OnPaint() 
{
	//CPaintDC dc(this); // device context for painting
	
	// TODO: Add your message handler code here
	
	// Do not call CDialog::OnPaint() for painting messages
	Invalidate();
	RedrawWindow(NULL,NULL,RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
	if ( m_bGrabbedImgDlg ) {
		m_pGrabbedImgDlg->PostMessage( WM_PAINT );
	}
	m_picProcessedImg.PostMessage( WM_PAINT );
}

void CHistogramDialog::PostNcDestroy() 
{
	// TODO: Add your specialized code here and/or call the base class
	CleanUp();
	CDialog::PostNcDestroy();
}

void CHistogramDialog::OnRobust() 
{
	// TODO: Add your control notification handler code here
	// Determines whether the histogram type is robust or fast
	UpdateData(TRUE);
}
