/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

///////////////////////////////////////////////////////////////////////////
// Filename	: OdxIsr.cpp
// Created	: Wednesday, May 14th, 1997
// Description: Implementation of the OdxIsr routines
// 
//  Vasanth Philomin 1997, vasi@cs.umd.edu
///////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "OdxIsr.h"
#include <afxmt.h>	// For multithreading and synchronization

// Globals
static CEvent stopEvent(FALSE, FALSE, "ODX Stop Event");		// AutoReset event that needs to be set to remove the profile handler i.e.
// exit the thread
// Define all the counts as 32-bit unsigned integers
// There was no choice but to define these counts as globals - the use of static atleast protects
// them beyond this file
UINT nInterrupts = 0;		// Number of Interrupts received from the Odx device
static UINT nOddFields = 0;			// Number of Odd Fields Acquired
static UINT nEvenFields = 0;		// Number of Even Fields Acquired
UINT nFrames = 0;			// Number of Frames Acquired
UINT nFields = 0;			// Number of Fields Acquired
static UINT nEndOfGrabs = 0;		// Number of Times the Acquisition has stopped.
static HWND hWnd = 0;				// Handle to the window to which interrupt messages are to be sent


void SetMsgWindowHandle(HWND hWindow)
/*
Description:
This function should be called to set the window to which interrupt
messages are to be sent.
Arguments:
hWindow - Handle to the window to which messages are to be sent
Returns:
Void
*/
{
	hWnd = hWindow;
}

void OdxInstallProfileHandler()
{
	// Start the ProfileHandler Thread
	AfxBeginThread(OdxProfileHandlerThread, 0);
}

void OdxRemoveProfileHandler()
{
	stopEvent.SetEvent();
}

BOOL OdxIsr(DWORD event)
{
	// Count number of interrupts
	nInterrupts++; 
	
	if (event & EVENTMODE_END_OF_FRAME)	{
		nFrames++;
		// Send User defined message to main application
		if (hWnd != NULL) 
			PostMessage(hWnd, WM_ODXISR, nFrames, (LPARAM)event);
	}
	else if (event & EVENTMODE_START_OF_FIELD)
		nFields++;
	
	return TRUE;	// To indicate that the event has been handled
}

UINT OdxProfileHandlerThread(LPVOID pParam)
{
	seteventfct(OdxIsr);
	
	//	::WaitForSingleObject(stopEvent.m_hObject, INFINITE);	// Wait for the stop event to be signalled
	
	//	seteventfct(NULL);
	return 0;	// Ends the thread
}

void ResetCounters()
{
	nInterrupts = nFrames = nFields = nEndOfGrabs = nEvenFields = nOddFields = 0;
}
