// File       : UAdaptEM.cpp
// Creation   : Feb 97
// Creator    : N.OLIVER
// Goal       : TAdaptEM class definition. See "UAdaptEM.h".
//
// Tab        : yes

#include "stdafx.h"
// unit include
#include "UAdaptEM.h"
#include "matmath.h"
#include "UStatistics.h"

// project includes
#include "UObjAlloc.h"

////////// defines
#define SIEVEBOUND     1000.
#define INITVARWEIGHT  10.  
#define MAX_EXP        40.
#define MIN(a,b)       (((a)<(b))?(a):(b))
#define MAX(a,b)       (((a)<(b))?(b):(a))
#define MIN_FLOAT      .01
#define SIGN(a)        ((a)>0 ? (1):(-1))

////////// static data members initialisation


//
//    Init
//

ErrorNumber TAdaptEM::Init ( TMixGaussMutex  *m,
			     TBitMap         *source,
			     TROIMutex       *roi,
			     TFlagMutex      *flag,
			     UINT           maxNumComp,
			     TGenGaussian    *models,
			     UINT           numComp,
			     float           *weights,
			     float            initDist,
			     UInt16           dim,
			     Boolean          interior,
			     float*           margin,
			     float            updateTime,
			     double           Tc,	      
			     double           initVar,
			     int              numIter,
			     char*            fName )
{
  ErrorNumber       err = kErrNoErr;
  int               i;
  // init data members
  m_nMaxTerms = maxNumComp;
  // default maximum number of terms
  m_nNumTerms = 0;
  m_nNumPts = 0;
  m_nNumObs = 0;
  m_nInitNumTerms = 0;
  m_bInitModelsFlag = m_bUpdateML = false;
  m_fFactorProb = m_fInitFactorProb = m_fDensity = 0.0;
  m_fMaxDist = m_fLastMaxDist = HUGE;
  fTimer = (TTimer*)NULL;

  m_pInitModels = (TGenGaussian*)NULL;
  m_pInitW = (double*)NULL;

  m_bInterior = interior;
  bzero( m_arrMargin, sizeof(float)*2 );
  if ( margin != (float*)NULL )
    {
      m_arrMargin[0] = margin[0];
      m_arrMargin[1] = margin[1];
    }

  m_pW = (double*)NULL;
  m_pModels = (TGenGaussian*)NULL;
  m_pGaussOutputs = (double*)NULL;
  m_pScaledOutputs = (double*)NULL;
  m_pAverageDist = (double*)NULL;
  m_pInputData = (double*)NULL;

  m_pW = new double[m_nMaxTerms];
  m_fInitVar = initVar;
  m_pModels = new TGenGaussian[m_nMaxTerms];
  m_pGaussOutputs = new double[m_nMaxTerms];
  m_pScaledOutputs = new double[m_nMaxTerms];
  m_pAverageDist = new double[m_nMaxTerms];
  m_fUpdateTime = updateTime;
  m_nNumIter = numIter;// only one iteration
  m_nMaxObs = 15000;   // maximum number of observations

  m_fTc = Tc;
  m_nDim = dim;
  // Initialize data members
  bzero(m_pW, sizeof(double)*m_nMaxTerms);
  bzero(m_pScaledOutputs, sizeof(double)*m_nMaxTerms);
  bzero(m_pGaussOutputs, sizeof(double)*m_nMaxTerms);
  bzero(m_pAverageDist, sizeof(double)*m_nMaxTerms);
  m_fAverageLike = 0;
  m_fInitAverageLike = 0;
  m_fInitNormLike = 0;
  m_pInputData = new double[m_nMaxObs*m_nDim];

  for (i=0;i<m_nMaxTerms;i++)
  m_pModels[i].Init(m_nDim);
  
  if ( models != NULL )
  {
    m_nNumTerms = numComp;
    m_nInitNumTerms = numComp;
    m_pInitModels = new TGenGaussian[ m_nNumTerms ];
    m_pInitW = new double[ m_nNumTerms ];

    for (i=0;i<m_nNumTerms;i++)
    {
      m_pModels[i].set(models[i]);
      m_pInitModels[i].Init(m_pModels[i]);
      m_pInitModelsFlag = true;
      if ( weights != NULL )
      {
	m_pW[i] = m_pInitW[i] = (double)weights[i];
      } else
      {
	m_pW[i] = m_pInitW[i] = (double)1./(double)m_nNumTerms;
      }
      if ( initDist != 0 )
	fInitMaxDist = initDist;
   }
  }
  // output stream
  if ( fName != NULL )
    {
      fOutputStream = new(ofstream);
      fOutputStream->open(fName);
      if ( fOutputStream->rdstate() != 0 ) 
	err = kErrFileErr;
    }


  err = AllocAndInit (fTimer);
  // error handling
  if (err != kErrNoErr)
  this->Dispose ();

  fTimer->Gap();

  // call inherited::Init
  err = TLearn::Init ( m, source, roi, flag, interior );

  // error handling
  if (err != kErrNoErr)
  this->Dispose ();

  fTimeElapsed = 0;
  fFirstTime = true;
  //  cout << "Initial update time " << m_fUpdateTime << endl;
  return err;
}




//
//    Dispose
//

void TAdaptEM::Dispose ()
{

  // Dispose data members
  if ( m_pW != (double*)NULL )
    delete m_pW;
  m_pW = (double*)NULL;

  for (int i=0; i<m_nMaxTerms; i++)
      m_pModels[i].Dispose();

  if ( m_pModels != (TGenGaussian*)NULL )
    delete(m_pModels);
  m_pModels = (TGenGaussian*)NULL;

  if ( m_pAverageDist != (double*)NULL )
    delete(m_pAverageDist);
  m_pAverageDist = (double*)NULL;

  if ( m_pGaussOutputs != (double*)NULL )
    delete(m_pGaussOutputs);
  m_pGaussOutputs = (double*)NULL;

  if ( m_pScaledOutputs != (double*)NULL )
    delete(m_pScaledOutputs);
  m_pScaledOutputs = (double*)NULL;
  if ( m_pInputData != (double*)NULL )
    delete(m_pInputData);
  m_pInputData = (double*)NULL;

  if ( m_pInitModels != NULL )
    delete(m_pInitModels);
  m_pInitModels = (TGenGaussian*)NULL;

  if ( m_pInitW != NULL )
    delete(m_pInitW);
  m_pInitW = (double*)NULL;

  if ( fOutputStream != NULL )
    {
      fOutputStream->close();
      delete(fOutputStream);
    }
  fOutputStream = (ofstream*)NULL;


  // dispose the timer
  TestDisposeAndFree (fTimer);
  // call inherited::Dispose
  TLearn::Dispose ();

}


//
//    Reads data and saves them in m_pInputData
//

ErrorNumber
TAdaptEM::ReadData()
{
  UINT         offset;
  UINT         *basePtr, *linePtr, *pixPtr;       // bitmap pointers
  UINT         r, g, b, lum;                     // RGB components
  int            i, j,width, height, initW, endW, initH, endH;
  ErrorNumber    err = kErrNoErr;
  CRect          m_pBB;
  float          m_pBBW, m_pBBH;

  m_nNumObs = 0;
  // Read data points and include them in the model
  err =  InitProcess( width, height, offset, basePtr, m_pBB, m_fDensity );

  if (err != kErrNoErr)
  return err;
  
  if ( m_bInterior )
    {
      for ( i=0; i<height/2; i++ ) 
	{
	  for ( j=0; j<width/2;j++ )
	    {
	      pixPtr=basePtr+i*offset+j*4;
	      pixPtr++;
	      b = *(pixPtr);
	      pixPtr++;
	      g = *pixPtr;
	      pixPtr++;
	      r = *pixPtr;
	      pixPtr++;
	      // compute pixel luminance
	      lum = r+g+b;
	      m_pInputData[m_nNumObs*m_nDim] = ((lum > 0 ) ? (double)r/(double)lum : 0);
	      m_pInputData[m_nNumObs*m_nDim+1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
	      m_nNumObs++;
	      if ( j>0 )
		{
		  pixPtr=basePtr+i*offset-j*4;
		  pixPtr++;
		  b = *pixPtr;
		  pixPtr++;
		  g = *pixPtr;
		  pixPtr++;
		  r = *pixPtr;
		  pixPtr++;
		  // compute pixel luminance
		  lum = r+g+b;
		  m_pInputData[m_nNumObs*m_nDim] = ((lum > 0 ) ? (double)r/(double)lum : 0);
		  m_pInputData[m_nNumObs*m_nDim+1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
		  m_nNumObs++;
		  if ( i>0 )
		    {
		      pixPtr=basePtr-i*offset-j*4;
		      pixPtr++;
		      b = *pixPtr;
		      pixPtr++;
		      g = *pixPtr;
		      pixPtr++;
		      r = *pixPtr;
		      pixPtr++;
		      // compute pixel luminance
		      lum = r+g+b;
		      m_pInputData[m_nNumObs*m_nDim] = ((lum > 0 ) ? (double)r/(double)lum : 0);
		      m_pInputData[m_nNumObs*m_nDim+1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
		      m_nNumObs++;
		    }
		} else if ( i>0 )
		  {
		    pixPtr=basePtr-i*offset+j*4;
		    pixPtr++;
		    b = *pixPtr;
		    pixPtr++;
		    g = *pixPtr;
		    pixPtr++;
		    r = *pixPtr;
		    pixPtr++;
		    // compute pixel luminance
		    lum = r+g+b;
		    m_pInputData[m_nNumObs*m_nDim] = ((lum > 0 ) ? (double)r/(double)lum : 0);
		    m_pInputData[m_nNumObs*m_nDim+1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
		    m_nNumObs++;
		  }
	    } // for j
	} // for i      
    } else 
      {
	// points to the left top corner
	linePtr = basePtr;
	m_pBBW = m_pBB.right-m_pBB.left;
	m_pBBH = m_pBB.bottom-m_pBB.top;
	if ( m_arrMargin[0] > 0 )
	  {
	    initW = MAX( (m_pBB.left-m_pBBW*m_arrMargin[0]), 0 );
	    endW = MIN( (m_pBB.right+m_pBBW*m_arrMargin[0]), width );
	  } else 
	    {
	      initW = 0;
	      endW = width;
	    }
	if  ( m_arrMargin[1] > 0 )
	  {
	    initH = MAX( (m_pBB.top-m_pBBH*m_arrMargin[1]), 0 );
	    endH = MIN( (m_pBB.bottom+m_pBBH*m_arrMargin[1]), height );
	  } else 
	    {
	      initH = 0;
	      endH = height;
	    }
	//cout << "Limits of exterior " << initW  << " " << endW << " " 
	//     << initH << " " << endH << endl;
	// Add to the pointer the lines and columns of the first location
	linePtr += initH*offset + initW*4;
	// read the exterior of the m_pBB
	for ( i=initW; i<endW ; i++ )
	  {
	    pixPtr = linePtr;
	    for ( j=initH; j<endH; j++ )
	      {
		if ( i>m_pBB.left && i<m_pBB.right && j>m_pBB.top && j<m_pBB.bottom )
		  break;
		pixPtr++;
		b = *pixPtr;
		pixPtr++;
		g = *pixPtr;
		pixPtr++;
		r = *pixPtr;
		pixPtr++;
		// compute pixel luminance
		lum = r+g+b;
		m_pInputData[m_nNumObs*m_nDim] = ((lum > 0 ) ? (double)r/(double)lum : 0);
		m_pInputData[m_nNumObs*m_nDim+1] = ((lum > 0 ) ? (double)g/(double)lum : 0);
		m_nNumObs++;
	      } // for j
	    linePtr += offset;
	  }// for i
      } // else
  return err;
}


//
//     GetGaussianOutputs
//
//   Get the values for the individual terms.  
//   The output is not updated by this routine.  
//   No learning (updating of terms) is done here either.
//   Get the value of every Gaussian term at fInput[] 
//   and store scaled and unscaled values.

ErrorNumber
TAdaptEM::GetGaussianOutputs()
{
  ErrorNumber      err  = kErrNoErr;  
  int              i,j,k;
  double           mahal2, det, gaussOutput;
  double*          m;
  double*          ic;
  double           initOutput=0.0;

  fOutput = 0.0;
  for(i=0;i<m_nNumTerms;i++)
    {
      if(m_pW[i] > 0.0)
      {
	mahal2=0.0;
	for(j=0;j<m_nDim;j++)
	{
	  for(k=0;k<m_nDim;k++)
	  {
	    m=m_pModels[i].getMean();
	    ic=m_pModels[i].getInvCov(); 
	    mahal2 += (double)(m[j] - fInput[j])*(double)(m[k] - fInput[k])*
					 (double)ic[j*m_nDim+k];
	  } // for k
	} // for j

	if( fabs(mahal2) > MAX_EXP)
	  {
	    m_pGaussOutputs[i] = 0.0;
	  }
	else
	  {
	    m_pGaussOutputs[i] = exp(-.5*mahal2);
	  }
	det = m_pModels[i].getDet();
	if ( det > 0 )
	  {
	    m_pScaledOutputs[i] = 
	      m_pGaussOutputs[i]/(pow(2.*M_PI,m_nDim/2.)*sqrt(det));
	  } else 
	    {
	      m_pScaledOutputs[i] = 0.0;
	    }
	fOutput += m_pW[i]*m_pScaledOutputs[i];
      } else
      {
	m_pScaledOutputs[i] = m_pGaussOutputs[i] = 0.0;
      }
    } // for
  if ( fOutput )
    m_fAverageLike += log(fOutput);

  for(i=0;i<m_nInitNumTerms;i++)
    {
      mahal2=0.0;
      for(j=0;j<m_nDim;j++)
	{
	  for(k=0;k<m_nDim;k++)
	    {
	      m=m_pInitModels[i].getMean();
	      ic=m_pInitModels[i].getInvCov(); 
	      mahal2 += (double)(m[j] - fInput[j])*(double)(m[k] - fInput[k])*
					 (double)ic[j*m_nDim+k];
	    } // for k
	} // for j
      
      if( fabs(mahal2) < MAX_EXP)
	{
	  gaussOutput = exp(-.5*mahal2);
	  det = m_pInitModels[i].getDet();
	  if ( det > 0 )
	    {
	      gaussOutput /= (pow(2.*M_PI,m_nDim/2.)*sqrt(det));
	    } else 
	      {
		gaussOutput = 0.0;
	      }
	  initOutput += m_pInitW[i]*gaussOutput;
	} 
    } // for
  if ( initOutput > 0 )
    m_fInitAverageLike += log(initOutput);

  return(err);
}

//
//      AddTerm
//

ErrorNumber
TAdaptEM::AddTerm()
/* Create a new term.                   */
/* This is C() in Equation (4.6),       */
/* defined by Equations (3.12) - (3.16) */
{
  ErrorNumber       err = kErrNoErr;
  int               i,j,k;
  double            tmp;
  double            tvar=0.0;
  double            *tcov;
  double            totalscaled=0.0;
  int               usetvar=0;
  double            *auxP, *auxP1;

  if(m_nNumTerms >= m_nMaxTerms)
  {
    fprintf(stderr,"\n%d term",m_nNumTerms);
    fprintf(stderr,"Overflow:  m_nMaxTerms.\n");
    fprintf(stderr,"Cannot create new term.\n");
    exit(1);
  } 

/* Determine tvar from existing term. */

/* find total output of all terms */
  for(i=0;i<m_nNumTerms;i++) totalscaled+=m_pScaledOutputs[i];

/* if total output is zero, then use default */
  if(totalscaled==0.0 || m_nNumTerms == 0) 
  {
    tvar = m_fInitVar;
    usetvar = -1;
  }
/* else make the variance the weighted average 
   of all other terms */
  else 
  {
    usetvar = 0;
    tcov = new double[ m_nDim*m_nDim ];
    for(i=0;i<m_nNumTerms;i++) 
    {
      for(j=0;j<m_nDim;j++) 
      {
	for(k=0;k<m_nDim;k++) 
	{
	  tcov[j*m_nDim+k] += (m_pScaledOutputs[i]/totalscaled)*m_pModels[i].getElemCov(j*m_nDim+k);
	}
      }
    } // for i
  } // else

/* Now set up the new term. */
  
  for(i=0;i<m_nDim;i++)
  {
    m_pModels[m_nNumTerms].setElemMean( fInput[i], i);
    if(usetvar)
    {
      m_pModels[m_nNumTerms].setElemCov( tvar, i*m_nDim+i );
    }
    else
    {
      for(j=0;j<m_nDim;j++)
      m_pModels[m_nNumTerms].setElemCov( tcov[i*m_nDim+j], i*m_nDim+j );
    }
  } // for i
  if (m_nNumPts>0)
    m_pW[m_nNumTerms] = 1.0/m_nNumPts;
  auxP = m_pModels[m_nNumTerms].getCov();
  auxP1 = m_pModels[m_nNumTerms].getInvCov();
  if (m_nDim==2)
  {
    DMatInvert2(auxP,&tmp,auxP1);
  } else
  {
    DMatInvert(auxP,m_nDim,&tmp,auxP1);
  }
  m_pModels[m_nNumTerms].setDet( tmp );

  m_nNumTerms++;

  if(!usetvar) delete(tcov);
  return(err);
}

//
//     UpdateW
//
ErrorNumber
TAdaptEM::UpdateW(int    term,
		 double amt)
/* Update the mixture percentage of the term */
{
  ErrorNumber       err = kErrNoErr;
  if (m_nNumPts>1)
  m_pW[term] += (amt - m_pW[term])/(m_nNumPts-1.);
  return(err);
}
/*****************************************************/

/*****************************************************/
ErrorNumber
TAdaptEM::UpdateMean( int    term,
		      double amt)
/* Update the mean of the term */
{
  ErrorNumber      err = kErrNoErr;
  int              i;
  double          *aux = m_pModels[term].getMean();

  if ((m_nNumPts>1) && m_pW[term]>0)
    {
      for(i=0;i<m_nDim;i++, aux++)
	{
	  (*aux) += (amt/((m_nNumPts-1.)*m_pW[term])) *
	    (fInput[i] - (*aux));
	}
    }
  return(err);
}

/*****************************************************/

/*****************************************************/
ErrorNumber
TAdaptEM::UpdateCov(int    term,
		    double amt)
/* Update the var of the term. */
{
  ErrorNumber      err = kErrNoErr;
  int              i,j;
  double           tmp, *auxP, *auxP1;
  double*          ptrCov = m_pModels[term].getCov();

  if ( (m_nNumPts>1) && (m_pW[term]>0) )
    {
      for(i=0;i<m_nDim;i++)
	{
	  for(j=0;j<m_nDim;j++, ptrCov++)
	    {
	      (*ptrCov) += 
		(amt/((m_nNumPts-1.)*m_pW[term]+INITVARWEIGHT)) *
		( (m_pModels[term].getElemMean(i) - fInput[i]) *
		  (m_pModels[term].getElemMean(j) - fInput[j]) - (*ptrCov) );
	      auxP = m_pModels[term].getCov();
	      auxP1 = m_pModels[term].getInvCov();
	      if ( m_nDim==2 )
		{
		  DMatInvert2(auxP,&tmp,auxP1);
		} else
		  {
		    DMatInvert(auxP,m_nDim,&tmp,auxP1);
		  }
	      m_pModels[term].setDet( tmp );
	    }
	} //for
    }
  return(err);
} /* end update cov */

//
//     LearnStep
//
//     On-line learning of the next sample
//     D_new = [1-P(zeta ; D_old)] U(zeta ; D_old) + 
//              P(zeta ; D_old)  C(zeta ; D_old)

ErrorNumber
TAdaptEM::LearnStep()
{
  ErrorNumber      err  = kErrNoErr;  
  int i,j;
  int P=1;
  double amt;
  double maxv=0.0;

  if(m_nNumTerms == 0){
    AddTerm();
    return err;
  }
  if(m_nNumTerms < m_nMaxTerms){
    for(i=0;i<m_nNumTerms;i++)
      if(m_pGaussOutputs[i] > maxv )
	maxv = m_pGaussOutputs[i];
    /* create decision */
    if( m_fTc <= maxv )
      {
	P=0;
      }
  } /* end if numterms */
  else
    {
      P=0;
    }
  if (P==1) 
    // C()
    {
      AddTerm();
    }
  else 
    {  
      // U
      
      for(i=0;i<m_nNumTerms;i++){
      
	if(fOutput > 0.0) 
	  {
	    amt = m_pScaledOutputs[i]*m_pW[i]/fOutput;
	  }
	else 
	amt = 0.0;

	UpdateW(i,amt);
	if(amt > 0.0){
	    UpdateCov(i,amt);
	    UpdateMean(i,amt);
	  } /* end if amt */

      } /* end for */
  } /* end else (not P) */
  return(err);
}

//
//     Normalize
//
ErrorNumber
TAdaptEM::Normalize()
{
  ErrorNumber      err  = kErrNoErr;
  int i,j,k;
  double tmp;
  double *ptrTmp, *auxP, *auxP1;
  double sum=0.0;
  double under[1000],over[1000];
  double tunder=0.0,tover=0.0,piprime;

  for(i=0;i<m_nNumTerms;i++) sum += m_pW[i];
  for(i=0;i<m_nNumTerms;i++) m_pW[i] = m_pW[i] / sum;

  /* Implement bounds dictated by sieve S4 */
  /* Equation (4.4)                        */

  /* Bounds on p eliminated.  Uncomment if desired
     piprime=1./(SIEVEBOUND*numterms);
  */
  piprime=0.0;

  for(i=0;i<m_nNumTerms;i++)
  {
    if(m_pW[i]<piprime)
    {
      under[i]=piprime-m_pW[i];
      tunder+=piprime-m_pW[i];
      over[i]=0.0;
    }
    else under[i]=0.0;

    if(m_pW[i]>piprime){
	over[i]=m_pW[i]-piprime;
	tover+=m_pW[i]-piprime;
	under[i]=0.0;
      }
    else over[i]=0.0;
  }
  
  for(i=0;i<m_nNumTerms;i++)
  {
    if(under[i]>0.0) m_pW[i]=piprime;
    if(over[i]>0.0) m_pW[i]-=tunder*(over[i]/tover);
    if(m_pModels[i].getDet() <= 0.0)
    {
      for(j=0;j<m_nDim;j++)
      {
	m_pModels[i].setElemCov(0.0, j*m_nDim+j );
      }
    }
    for(j=0;j<m_nDim;j++)
    {
      ptrTmp = m_pModels[i].getMean()+j;
      if(*ptrTmp>SIEVEBOUND*m_nNumTerms) 
      {
	(*ptrTmp)=SIEVEBOUND*m_nNumTerms;
      }
      if((*ptrTmp)<-SIEVEBOUND*m_nNumTerms)
      {
	(*ptrTmp)=-SIEVEBOUND*m_nNumTerms;
      }
      ptrTmp = m_pModels[i].getCov()+j*m_nDim;
      if(*(ptrTmp+j)>SIEVEBOUND*m_nNumTerms) 
      {
	for(k=0;k<m_nDim;k++)
	{
	  *(ptrTmp+k) = 0.0;
	}
	*(ptrTmp+j) = SIEVEBOUND*m_nNumTerms;
      }
      if(*(ptrTmp+j)<1./(SIEVEBOUND*m_nNumTerms)) 
      {
	for(k=0;k<m_nDim;k++)
	{
	  *(ptrTmp+k) = 0.0;
	}
	*(ptrTmp+j) = 1./(SIEVEBOUND*m_nNumTerms);
      }
    } // for j
    auxP = m_pModels[i].getCov();
    auxP1 = m_pModels[i].getInvCov();
    if ( m_nDim == 2 )
    {
      DMatInvert2(auxP,&tmp,auxP1);
    } else
    {
      DMatInvert(auxP,m_nDim,&tmp,auxP1);
    }
    m_pModels[i].setDet(tmp);
  } // for i
  
  return(err);
} // normalize

//
//     Feed Data
//

ErrorNumber
TAdaptEM::FeedData()
{
  ErrorNumber      err  = kErrNoErr;
  int              i;

  //  cout << "Before FeedData "<< endl;
  //  for ( i=0; i<m_nNumTerms; i++ )
  //  m_pModels[i].print(cout);

  for ( i=0; i<m_nNumObs; i++ )
  {
    fInput = m_pInputData+(i*m_nDim);
    m_nNumPts++;
    GetGaussianOutputs();
    LearnStep();
    Normalize();
  }
  //cout << fInterior << " After FeedData "<< endl;
  //cout << "AverageLike " << m_fAverageLike << endl;
  //cout << "InitAverageLike " << m_fInitAverageLike << endl;
  //cout << "InitMaxDist " << fInitMaxDist << endl;
  //cout << "InitNormLike " << m_fInitNormLike << endl;

  if ( m_fAverageLike == 0 )
    {
      m_nNumTerms = 0;
    } else 
      {
	/* for ( i=0; i<m_nNumTerms; i++ )
	  {
	    cout << " weight " << m_pW[i];
	    m_pModels[i].print(cout);
	  } */
      }
  return(err);
}

//
//     Learn using adaptive mixture of gaussians
//
ErrorNumber
TAdaptEM::Learn()
{
  ErrorNumber      err  = kErrNoErr;
  int              i,j,index,localFlag, validTerms=0;
  float           *weights;
  TGaussian       *modelsArray;
  float            auxMean[2];
  float            auxCov[3];
  float            cumSumm_pW=0.0;

  // read the data every X seconds
  if ( ((fTimeElapsed/kTimerNbTiksPerSecond)>m_fUpdateTime) || (fFirstTime == true) )
  {
    //    cout << " Time elapsed " << fTimeElapsed << " m_fUpdateTime " << m_fUpdateTime << endl;
    fFirstTime=false;
    fTimeElapsed=0;
    err = ReadData();
    if ( (err == kErrNoErr) || (err == kErrInitModels) )
    {
      cout << "DENSITY " << m_fDensity << endl;
      if ( m_fDensity < .8 )
	{
	  for (i=0;i<m_nNumIter;i++)
	    {
	      if ( err == kErrNoErr )
		{
		  FeedData();
		}
	    }
	  if ( (err == kErrNoErr) && (m_nNumTerms>0) )
	    {
	      weights = new float[m_nNumTerms];
	      modelsArray = new TGaussian[m_nNumTerms];
	      for (i=0;i<m_nNumTerms;i++)
		{
		  if ( m_pW[i] < MIN_FLOAT )
		    continue;
		  if ( i != validTerms )
		    {
		      m_pModels[validTerms].set( m_pModels[i] );
		      m_pW[validTerms] = m_pW[i];
		      //		      m_pAverageDist[validTerms]=m_pAverageDist[i];
		    }
		  cumSumm_pW+=m_pW[validTerms];
	      //m_pAverageDist[validTerms]/=(m_nNumObs*m_nNumIter);
	  
	      for (j=0;j<2;j++)
		{
		  auxMean[j]=(float)m_pModels[validTerms].getElemMean(j);
		  auxCov[j]=(float)m_pModels[validTerms].getElemCov(j);
		}
	      auxCov[2]=(float)m_pModels[validTerms].getElemCov(3);
	  
	      modelsArray[validTerms].set(auxMean,
					  auxCov);
	  
	      weights[validTerms]=(float)m_pW[validTerms];
	      validTerms++;
	    } // for i=0  numTerms
	  if ( validTerms > 0 )
	    {
	      //cout << "m_fAverageLike " << m_fAverageLike << endl;
	      m_fLastMaxDist = m_fMaxDist;
	      m_fMaxDist=m_fAverageLike/(m_nNumPts*m_nNumIter);
	      cout << " Maxdist " << m_fMaxDist << " Last max dist " << m_fLastMaxDist << endl;
	      if ( ( m_pInitW != (double*)NULL ) )
		{
		  fInitMaxDist = m_fInitAverageLike/(m_nNumPts*m_nNumIter); 
		  
		  if (m_fInitNormLike == 0 )
		    {
		      m_fInitNormLike = fInitMaxDist;
		      m_fInitFactorProb = (m_fInitNormLike>0) ? .25 : 4;
		    }
		  cout << " Init max dist " << fInitMaxDist << endl;
		  cout << " Init normalized likelihood " << m_fInitNormLike << endl;
		}

	      // check validity of model
	      if ( ((fInitMaxDist < (m_fInitNormLike-fabs(m_fInitNormLike) ) ) || (m_fDensity<.5))
		   && (m_pInitModelsFlag == false) )
		{
		  // model is not correct
		  // load initial model
		  m_pOutputModels->mutex.SemP ();
		  m_pOutputModels->numGauss = m_nInitNumTerms;
		      for ( i=0; i<m_nInitNumTerms; i++ )
			{
			  for (j=0;j<2;j++)
			    {
			      auxMean[j]=(float)m_pInitModels[i].getElemMean(j);
			      auxCov[j]=(float)m_pInitModels[i].getElemCov(j);
			    }
			  auxCov[2]=(float)m_pInitModels[i].getElemCov(3);
			  
			  m_pOutputModels->gauss[i].set(auxMean, auxCov );
			  m_pOutputModels->w[i]=(float)m_pInitW[i];
			} // for i
		      m_fLastMaxDist = m_fInitNormLike;
		      m_pOutputModels->maxDist = m_fInitFactorProb*m_fInitNormLike;

		      m_pOutputModels->mutex.SemV ();
		      //cout << "Max dist " << m_fInitNormLike << endl;
		      // set flag to learn
		      if ( m_bLearnFlag != (TFlagMutex*)NULL )
			{
			  m_bLearnFlag->mutex.SemP ();
			  localFlag = m_bLearnFlag->flag = 1;
			  m_bLearnFlag->mutex.SemV ();
			}
		      cout << "Initialising models and the flag " << localFlag << endl;
		      
		      m_pInitModelsFlag = true;
		} else if ( (m_pInitModelsFlag == true ) &&
			    (fInitMaxDist < (m_fInitNormLike-1.5*fabs(m_fInitNormLike)) ) )
		  {
		    // set the m_fInitNormLike to fInitMaxDist
		    cout << "CHANGING m_fInitNormLike " << endl;
		    m_fInitNormLike = fInitMaxDist;
		  } else {
		      // not too far from initial models and the initial models are set
		  // Important: the number of validTerms needs to be <= maxNum of models
		  //if ( m_pInitModelsFlag == true )
		  //  {
		      m_pOutputModels->mutex.SemP ();
		      m_pOutputModels->numGauss = validTerms;
		      for ( i=0; i<validTerms; i++ )
			{
			  m_pOutputModels->gauss[i].set(modelsArray[i]);
			  m_pW[i] /= cumSumm_pW;
			  m_pOutputModels->w[i]=(float)m_pW[i];
			  //maxDist += m_pAverageDist[i];
			  if ( fOutputStream != (ofstream*)NULL )
			    (*fOutputStream) << m_pOutputModels->gauss[i] << m_pOutputModels->w[i] << endl;
			} // for i
		      m_fFactorProb = (m_fMaxDist>0) ? .7 : 1.3;
		      m_pOutputModels->maxDist = m_fFactorProb*m_fMaxDist;
		      m_pOutputModels->mutex.SemV ();
		      // check whether to learn or not depending
		      // on a drop of the likelihood per point
		      if (m_fLastMaxDist < HUGE ) 
			{
			  if ( (m_bUpdateML == true) ||
			       (m_fDensity < .6 ) || 
			      ((m_fMaxDist-m_fLastMaxDist) > .2*fabs(m_fLastMaxDist))
			   || (( m_pInitModelsFlag == true ) && 
			       ( m_fMaxDist > fInitMaxDist) ) )
			    {
			      m_bUpdateML = false;
			      // set flag to learn
			      if ( m_bLearnFlag != (TFlagMutex*)NULL )
				{
				  m_bLearnFlag->mutex.SemP ();
				  localFlag = ++(m_bLearnFlag->flag);
				  m_bLearnFlag->mutex.SemV ();
				  cout << "flag updated " << localFlag << endl;
				  cout << "Last max dist " << m_fLastMaxDist << " smaller than maxDist " << m_fMaxDist << endl;
				}   
			    }
			} else 
			  {
			    m_bUpdateML = true;
			  }
		      //cout << "Max dist " << m_fMaxDist << endl;
		      m_pInitModelsFlag = false;
		      // Set the initial likelihood to this
		      //m_fInitNormLike = fInitMaxDist;
		  }
	      //	}
	    } else if ( (m_pInitModels != (TGenGaussian*)NULL) &&
			(m_pInitModelsFlag == false) )
	      {
		m_pOutputModels->mutex.SemP ();
		m_pOutputModels->numGauss = m_nInitNumTerms;
		for ( i=0; i<m_nInitNumTerms; i++ )
		  {
		    for (j=0;j<2;j++)
		      {
			auxMean[j]=(float)m_pInitModels[i].getElemMean(j);
			auxCov[j]=(float)m_pInitModels[i].getElemCov(j);
		      }
		    auxCov[2]=(float)m_pInitModels[i].getElemCov(3);
		    
		    m_pOutputModels->gauss[i].set(auxMean, auxCov );
		    m_pOutputModels->w[i]=(float)m_pInitW[i];
		  } // for i
		m_fLastMaxDist = m_fInitNormLike;
		m_pOutputModels->maxDist = m_fInitFactorProb*m_fInitNormLike;
		m_pOutputModels->mutex.SemV ();
		//cout << "Max dist " << m_fInitNormLike << endl;
		// set flag to learn
		if ( m_bLearnFlag != (TFlagMutex*)NULL )
		  {
		    m_bLearnFlag->mutex.SemP ();
		    localFlag = m_bLearnFlag->flag = 1;
		    m_bLearnFlag->mutex.SemV ();
		    cout << "Initialising models and flag " << localFlag << endl;
		  }
		m_bInitModelsFlag = true;
	      } else // not Init models and not valid terms->error 
		{
		  validTerms = 0;
		}
	  delete modelsArray;
	  delete weights;
	  m_nNumTerms = validTerms;
	} else  
	  {
	    if ( m_bInitModelsFlag == false )
	      {
		// err = kInitModels -> m_pModels have the initial data
		m_pOutputModels->mutex.SemP ();
		m_pOutputModels->numGauss = m_nNumTerms;
		for ( i=0; i<m_nNumTerms; i++ )
		  {
		    for (j=0;j<2;j++)
		      {
			auxMean[j]=(float)m_pModels[i].getElemMean(j);
			auxCov[j]=(float)m_pModels[i].getElemCov(j);
		      }
		    auxCov[2]=(float)m_pModels[i].getElemCov(3);
		    
		    m_pOutputModels->gauss[i].set(auxMean, auxCov );
		    m_pOutputModels->w[i]=(float)m_pW[i];
		  }
		m_fLastMaxDist = fInitMaxDist;
		m_pOutputModels->maxDist = m_fInitFactorProb*fInitMaxDist;
		m_pOutputModels->mutex.SemV ();
		m_bInitModelsFlag = true;
		// set flag to learn
	      }
	    if ( m_bLearnFlag != (TFlagMutex*)NULL )
	      {
		m_bLearnFlag->mutex.SemP ();
		localFlag = m_bLearnFlag->flag = 1;
		m_bLearnFlag->mutex.SemV ();
		cout << "Initialising models and flag " << localFlag << endl;
	      }
	    
	    //cout << "Max dist " << fInitMaxDist << endl;
	  } 
      
      // Reset the points counter
      m_nNumPts=0;
      m_fAverageLike=0;
      m_fInitAverageLike=0;
	}
    } else
      {
	cout << "Error reading data" << endl;
	m_pOutputModels->mutex.SemP ();
	m_pOutputModels->numGauss = 0;
	m_pOutputModels->mutex.SemV ();
	m_nNumTerms = 0;
	if ( m_bLearnFlag != (TFlagMutex*)NULL )
	  {
	    // signal learner to learn update histogram 
	    m_bLearnFlag->mutex.SemP ();
	    localFlag = m_bLearnFlag->flag = 1;
	    m_bLearnFlag->mutex.SemV ();
	    cout << "Initialising flag with error reading data " << localFlag << endl;
	  }
      }
  }
  bzero(m_pAverageDist, sizeof(double)*m_nMaxTerms);
  
  fTimeElapsed+=fTimer->Gap();
  return(err);
}
  



////////////////////////////////////////////////////////////////////////////////
//                                                                            //
//                         protected function members                         //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

//
//    InitProcess
//

inline
ErrorNumber TAdaptEM::InitProcess (int&        oWidth,
				   int&        oHeight,
				   UINT&     oOffset,
				   UINT*      &oBase,
				   CRect&      oROI,
				   float&      oDens )
{
  UINT          c[2];
  TBitMapSpec     spec;
  int             i;

  // get source bitmap parameters
  spec      = m_pBm->GetSpec ();

  if ((spec.encoding >> 8) != 4)
    // can only manage 4 byte per pixel source bitmap
    return kErrCantHandleEncoding;

  oOffset   = spec.lineOffset;

  // defines the ROI
  // wait for mutex
  m_pBB->mutex.SemP ();
  if (m_pBB->roi == CRect(0,0,0,0))
  {
    cout << "NOTHING TO PROCESS" << endl;
    m_pBB->mutex.SemV();
    // Initialize models
    if ( m_nInitNumTerms )
      {
	for ( i=0; i<m_nInitNumTerms; i++ )
	  {
	    m_nNumTerms = m_nInitNumTerms;
	    m_pW[i] = m_pInitW[i];
	    m_pModels[i].set(m_pInitModels[i]);
	  }
	return kErrInitModels;
      }
    oDens=0;
    return kErrNoData;
  }
  else
  {
    oROI = m_pBB->roi;
    oDens = m_pBB->density;
  }
  m_pBB->mutex.SemV ();


  if (m_bInterior)
    {
      // size of the area to process
      oWidth  = oROI.right   - oROI.left;
      oHeight = oROI.bottom  - oROI.top;
      c[0]=(oROI.right + oROI.left)/2;
      c[1]=(oROI.top + oROI.bottom)/2;
      // Source bitmap initialization:
      //  oBase = (UINT *)(m_pBm->GetBase()) + oROItop * oOffset + oROIleft * 4;
      oBase = (UINT *)(m_pBm->GetBase()) + c[1] * oOffset + c[0] * 4;
    } else
      {
	oWidth = spec.width;
	oHeight = spec.height;
	// point to the top right corner
	oBase = (UINT *)(m_pBm->GetBase());
      }
  return kErrNoErr;
}

