// File       : ULearn.cpp
// Creation   : Feb 1997
// Creator    : N.OLIVER
// Goal       : defines TLearn virtual class. See "ULearn.h"
//
// Tab        : yes


// unit include
#include "ULearn.h"

// project includes
#include "UObjAlloc.h"

// system includes

//
//    Init
//

ErrorNumber TLearn::Init ( TMixGaussMutex*  m,
			   TBitMap*     source,
			   TROIMutex*   ROI,
			   TFlagMutex*  flag,
			   Boolean      interior )
{
  ErrorNumber   err = kErrNoErr;
  
  fbm             = source;
  fOutputModels   = m;
  fInterior       = interior;
  bb              = ROI;
  fLearnFlag      = flag;
  fParentID       = 0;
  fChildID        = 0;

  cout << "FInterior " << interior << endl;
  // Initialize the mutexes
  //  TLearn::findex->mutex.Init(1);
  //TLearn::bb->mutex.Init(1);

  if (err == kErrNoErr)
  {
    // setup the parent process ID
    fParentID = (ProcessID)getpid ();
    
    // set fChildID to be different of 0, so that the child won't exit
    // prematurely
    fChildID = 1;
    unsigned int inh = PR_SALL | PR_SETEXITSIG;
    // fork to have a process learning the histogram
    fChildID = (ProcessID)sproc (LearnChildMain, inh, (void*)this);
    
    if (fChildID == -1)
    {
      fChildID = 0;
      cout << "Error forking child : No more processes " << endl;
      err = kErrNoMoreProcess;
    }
  }
  // error handling
  if (err != kErrNoErr)
    this->Dispose ();

  return err;
}




//
//    Dispose
//

void TLearn::Dispose ()
{
  // tell the child process to stop
  if (fChildID != 0)
  {
    // wait for the child to signal it is finished
    while (fParentID != 0)
    {
      signal(SIGTERM,SIG_IGN);
      kill( fChildID, SIGTERM );
      cout << "killed child" << endl;
      fParentID = 0;
      fChildID = 0;
    }
  }

}


//
//    LowLevelProcess
//

void TLearn::LowLevelProcess ( )
{

  // Here the child will learn
  this->Learn( );
}




//
//    LearnChildMain
//

void LearnChildMain (void* userData)
{
  TLearn*            l = (TLearn*)userData;

  signal(SIGTERM,SIG_DFL);
  while (l->fChildID)
  {
    l->LowLevelProcess ( );
    sginap (1);
  }

  // tell the parent process we are finished
  l->fParentID = 0;
}
