/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

//
//       File : UObjAlloc.h
//       Goal : Generic object allocation / deallocation functions.
//
//    Main modifs :
//
//  Table of content :
//    Allocate
//    AllocAndInit
//    DisposeAndFree
//    TestDisposeAndFree
//    GetGlobalObject
//    ReleaseGlobalObject
//  Nuria Oliver 1997, nuria@media.mit.edu

#ifndef _UOBJALLOC_H
#define _UOBJALLOC_H

// project includes
#include "UErrors.h"

#define Delete(x) Free (x)
#define TestDelete(x) TestAndFree (x)


//
//    Allocate
//

template<class T>
inline
ErrorNumber Allocate (T* &object)
{
	ErrorNumber err = kErrNoErr;
	
	object = (T*)new T;
	
	if (object == (T*)NULL)
		err = kErrMemFull;
	
	return err;
}


//
//    AllocAndInit
//

template<class T>
inline
ErrorNumber AllocAndInit (T* &object)
{
	ErrorNumber err = kErrNoErr;
	
	err = Allocate (object);
	
	if (err == kErrNoErr)
		err = object->Init ();
	
	return err;
}


//
//    DisposeAndFree
//

template<class T>
inline
void DisposeAndFree (T* &object)
{
	object->Dispose ();
	delete object;
	object = (T*)NULL;
}



//
//    TestDisposeAndFree
//

template<class T>
inline
void TestDisposeAndFree (T* &object)
{
	if (object != (T*)NULL)
		DisposeAndFree (object);
}



//
//    Free
//

template<class T>
inline
void Free (T* &object)
{
	delete object;
	object = (T*)NULL;
}



//
//    TestAndFree
//

template<class T>
inline
void TestAndFree (T* &object)
{
	if (object != (T*)NULL)
		Free (object);
}


#endif
