/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

///////////////////////////////////////////////////////////////////////////
// Filename	: VDib.cpp
// Created	: Thursday, June 13th, 1997
// Description: The Dib Class implementation.
// 
//  Vasanth Philomin 1997, vasi@cs.umd.edu
///////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include "VDib.h"

BITMAPINFO_8  VDib::s_bmScreen;

void VDib::setBitmapFormat()
{
	// Get the screen format
	HDC dc = ::GetDC (NULL);
	CDC* pDC = CDC::FromHandle (dc);
	CPalette pal;
	pal.CreateHalftonePalette (pDC);
	CPalette* pOldPal = pDC->SelectPalette (&pal, TRUE);
	pDC->RealizePalette ();
	
	// Make a bitmap compatible with the screen format:
	CBitmap bm;
	bm.CreateCompatibleBitmap (pDC,1,1);
	
	// Get the data
	s_bmScreen.bmiHeader.biSize = sizeof (BITMAPINFOHEADER);
	s_bmScreen.bmiHeader.biWidth = 0;
	s_bmScreen.bmiHeader.biHeight = 0;
	s_bmScreen.bmiHeader.biPlanes = 1;
	s_bmScreen.bmiHeader.biBitCount = 0;
	s_bmScreen.bmiHeader.biCompression = BI_RGB;  // might change for BI_BITFIELDS
	s_bmScreen.bmiHeader.biSizeImage = 0;
	s_bmScreen.bmiHeader.biXPelsPerMeter = 0;
	s_bmScreen.bmiHeader.biYPelsPerMeter = 0;
	s_bmScreen.bmiHeader.biClrUsed = 0;
	s_bmScreen.bmiHeader.biClrImportant = 0;
	
	::GetDIBits (pDC->GetSafeHdc (), (HBITMAP)bm.GetSafeHandle (),
		0,0,NULL, s_bmScreen, DIB_RGB_COLORS);
	// Do it one more time to get the color table:
	::GetDIBits (pDC->GetSafeHdc (), (HBITMAP)bm.GetSafeHandle (),
		0,0,NULL, s_bmScreen, DIB_RGB_COLORS);
	
	// If this is a BI_BITFIELDS, change it back to BI_RGB.
	if (s_bmScreen.bmiHeader.biCompression == BI_BITFIELDS) {
		s_bmScreen.bmiHeader.biCompression = BI_RGB;
	}
	// Let's check top-down dibs:
	int retInt = 0;
	s_bmScreen.bmiHeader.biHeight = -1;
	short r = Escape (pDC->GetSafeHdc(), QUERYDIBSUPPORT, sizeof(BITMAPINFO), (const char*)&s_bmScreen,
		&retInt);
	
	if ((retInt & QDI_DIBTOSCREEN) == 0)
		s_bmScreen.bmiHeader.biHeight = 1;
	pDC->SelectPalette (pOldPal, TRUE);
	::ReleaseDC (NULL, dc);
}

BOOL VDib::Create(const BITMAPINFO_8 &bmInfo, BOOL bMatchScreen)
{
	if (bMatchScreen)	{	// Match to screen format
		m_bmHeader = VDib::s_bmScreen;	// Assumes that setBitmapFormat() has been called before
		
		m_bmHeader.bmiHeader.biHeight *= bmInfo.bmiHeader.biHeight;
		m_bmHeader.bmiHeader.biWidth = bmInfo.bmiHeader.biWidth;
		UINT wWidthBytes = (m_bmHeader.bmiHeader.biWidth * m_bmHeader.bmiHeader.biBitCount);
		wWidthBytes = ((wWidthBytes+31) & (~31))/ 8;
		
		m_bmHeader.bmiHeader.biSizeImage = wWidthBytes * abs (m_bmHeader.bmiHeader.biHeight);
	}
	else {
		m_bmHeader = bmInfo;
		m_bmHeader.bmiHeader.biHeight *= -1;
	}
	
	m_nLineOffset = (m_bmHeader.bmiHeader.biSizeImage/abs(m_bmHeader.bmiHeader.biHeight));
	m_rect.SetRect(0, 0, bmInfo.bmiHeader.biWidth, abs(bmInfo.bmiHeader.biHeight));
	
	// Make a palette in case necessary
	int nColors = VDib::GetNumClrTableEntries((const BITMAPINFO *)&m_bmHeader);
	if (nColors)	{
		// Make a gray scale color table
		LPRGBQUAD lpRGB = m_bmHeader.ColorTable();
		for (int i=0; i<nColors; i++) {
			lpRGB->rgbBlue = i;
			lpRGB->rgbGreen = i;
			lpRGB->rgbRed = i;
			lpRGB->rgbReserved = 0;
			lpRGB++;
		}
	}
	
	return TRUE;
}

void VDib::Draw (CDC* pDC, UINT x, UINT y)
{
	if (!m_pBits)
		return;
    ::StretchDIBits(pDC->GetSafeHdc(),
		x,							// Destination x
		y,							// Destination y
		m_bmHeader.bmiHeader.biWidth,				// Destination width
		abs (m_bmHeader.bmiHeader.biHeight),        // Destination height
		0,							// Source x
		0,							// Source y
		m_bmHeader.bmiHeader.biWidth,               // Source width
		abs (m_bmHeader.bmiHeader.biHeight),        // Source height
		m_pBits,					// Pointer to bits
		m_bmHeader,					// BITMAPINFO
		DIB_RGB_COLORS,								// Options
		SRCCOPY);									// Raster operator code (ROP)
}


// Protected static function
int VDib::GetNumClrTableEntries(const BITMAPINFO* pBmpInfo) 
{
    int iColors;
	
    ASSERT(pBmpInfo);
	
    switch (pBmpInfo->bmiHeader.biBitCount) {
	case 1:
        iColors = 2;
        break;
    case 4:
        iColors = 16;
        break;
    case 8:
        iColors = 256;
        break;
		// Anything bigger has no color table.
    default:
        iColors = 0;
        break;
    }
	
    if (pBmpInfo->bmiHeader.biClrUsed != 0)
        iColors = pBmpInfo->bmiHeader.biClrUsed;
	
    return iColors;
}