/***************************************************************************
*
* Copyright 1997 by  the  Massachusetts Institute  of Technology.   All
* rights reserved.
*
* Developed  by the Perceptual Computing Section
* at the Media Laboratory, MIT, Cambridge, Massachusetts.
*
* Permission to use, copy,  or modify this software and  its documentation
* for  educational  and  research purposes only and without fee  is hereby
* granted, provided  that this copyright notice and the original authors's
* names appear  on all copies and supporting documentation.  If individual
* files are  separated from  this  distribution directory  structure, this
* copyright notice must be included.  For any other uses of this software,
* in original or  modified form, including but not limited to distribution
* in whole or in  part, specific  prior permission  must be  obtained from
* MIT.  These programs shall not  be  used, rewritten, or  adapted as  the
*   basis  of  a commercial  software  or  hardware product  without first
* obtaining appropriate licenses  from MIT.  MIT. makes no representations
* about the suitability of this  software for any purpose.  It is provided
* "as is" without express or implied warranty.
*
**************************************************************************/

///////////////////////////////////////////////////////////////////////////
// Filename	: Vision.h
// Created	: Tuesday, June 17th, 1997
// Description: A class to do all the computer vision processing
// 
//  Nuria Oliver 1997, nuria@media.mit.edu
///////////////////////////////////////////////////////////////////////////
// Vision.h : header file
//
#ifndef _VISION_H
#define _VISION_H

#include "ImgWnd.h"
#include "StaticFrame.h"
#include <afxmt.h>
#include "UErrors.h"
#include "UBitMap.h"
#include "TciGrabber.h"

class CVision;
class CHistogramDialog;
class CStaticFrameDialog;
class CStaticFrameBBDialog;

struct VisionThreadParams	{
	CVision		*pVision;	
	CStaticFrameBBDialog	*pEstim;
	CHistogramDialog	*pHisto;
	CStaticFrameDialog	*pConnex;
};

class CVision	{
	friend UINT VisionThread(LPVOID param);
public:
	
	ErrorNumber  Init( VTciGrabber  *grabber,
		int factor );
	void StartVision(	CStaticFrameBBDialog   *pEstim,
								 CHistogramDialog		 *pHisto,
								 CStaticFrameDialog		 *pConnex );
	void StopVision();
	void Dispose();
	
	// Get functions
	int		Factor() { return m_dFactor;}
	BOOL VisionOn() { return m_bVisionOn;}
	
	CVision();
	~CVision();
	
	CRITICAL_SECTION	m_Sem;
protected:
	CWinThread* pTh;
	VisionThreadParams			params;
	CEvent stopEvent;
	HANDLE					 m_Thread;
	TBitMap					   m_bm; 
	TBitMap                 m_histoBM;
	TBitMap                 m_connexBM;	
	VTciGrabber				*m_pGrabber;
	int						m_dFactor;
	BOOL					m_bVisionOn;
public:
	
};

#endif
